/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <sstream>
#include <iostream>
#include <string>

#include "spakes.h"
#ifdef WINDOWS
#include <winsock.h>
#include "..\client\protocol.h"
#include "..\client\point.h"
#else
#include "../client/protocol.h"
#include "../client/point.h"
#include <sys/socket.h>
#include <unistd.h>
#endif

#include "socket.h"
#include "player.h"
#include "init.h"


#define VERBOSE

using namespace std;

int maxplayers = 1;
int numplayers = 0;
Socket clientfd[4];
Player players[4];


void broadcast_cmd(int playernum, ostringstream &o, bool includesender = false)
{
	string scommand = o.str();

	for (int i = 0; i < maxplayers; i++) {
		if ( (i != playernum) || includesender) {
			if (players[i].ingame && players[i].primary) {
				send(clientfd[i].sockfd, scommand.c_str(), scommand.size(), 0);
			}
		}
	}
}

void ByeByePlayer(int pnum)
{
	closesocket(clientfd[pnum].sockfd);
	players[pnum].ingame = false;
	ostringstream playerout;
	playerout << SRV_PLAYER_OUT << ' ' << pnum << DELIMINATER;
	cout << "Player " << pnum + 1 << " left the game. " << endl;
	broadcast_cmd(pnum, playerout);

	int secplayer = players[pnum].secondarynum;
	if (secplayer > -1) {
		ostringstream secout;
		secout << SRV_PLAYER_OUT << secplayer << DELIMINATER;
		players[secplayer].ingame = false;
		broadcast_cmd(pnum, secout);
	}
}

void main_game_loop()
{

#ifdef VERBOSE
	cout << "Now entering main game loop" << endl;
#endif
	fd_set allset;
	while (true) {
		FD_ZERO(&allset);
		int maxfd = 0;
		for (int i = 0; i < numplayers; i++) {
			if (players[i].primary && players[i].ingame) {
				FD_SET(clientfd[i].sockfd, &allset);
				if (clientfd[i].sockfd > maxfd) maxfd = clientfd[i].sockfd;
			}
		}
		if (maxfd == 0) return;

		select(maxfd + 1, &allset, NULL, NULL, NULL);
		for (int j = 0; j < numplayers; j++) {
			if (FD_ISSET(clientfd[j].sockfd, &allset)) {
				string scommand;
				int numrecv;
				numrecv = clientfd[j].Readline(scommand);
				if ( numrecv < 0) {
					closesocket(clientfd[j].sockfd); return;
				} else if (numrecv == 0) {
					ByeByePlayer(j);
					scommand = "";
					continue;
				}

//				cout << "Data from player " << j << ": " << scommand << endl;

				istringstream command(scommand);
				int playerid, com_x, com_y, newx, newy;
				unsigned char com_comid;
				command >> com_comid;
				ostringstream reply;
				switch(com_comid) {
				case CMD_HITWALL:
					command >> playerid;
//					cout << "Player " << playerid << " lost a life!" << endl;
					reply << SRV_HITWALL << playerid << DELIMINATER;
					broadcast_cmd(j, reply);
					players[playerid].lose_life();
					break;
				case CMD_ATEAPPLE:
					command >> playerid;
					if (!players[playerid].Living()) break;

					command >> com_x >> com_y >> newx >> newy;
//					cout << "Got CMD_ATEAPPLE from: " << j << " (" << com_x << "/" << com_y << ")" << endl;
					ostringstream ateapple;
					ateapple << SRV_ATEAPPLE << playerid << ' ' << com_x << ' ' << com_y << ' ' << newx << ' ' << newy << ' ' << DELIMINATER;
					broadcast_cmd(j, ateapple, true);

					break;
				}
				scommand = "";
			}
		}
	}
}

int main(int argc, char **argv)
{
#ifdef WINDOWS
	WSADATA wsadata;
	WSAStartup(MAKEWORD(1,1), &wsadata);
#endif
	cout << "SPAKES Server " << VERSION << " by Markus Triska dedicated to S.P. -- Happy Birthday!" << endl;
	if (argc > 1) {
		maxplayers = atoi(argv[1]);
	} else {
		cout << "Please enter desired number of players: ";
		cin >> maxplayers;
	}

	if (maxplayers > 4) maxplayers = 4;
	if (maxplayers < 1) maxplayers = 1;

	cout << "OK, I'll wait for " << maxplayers << " players!" << endl << endl;


	while(true) {
		if (init::establish_connections()) {
			if (init::initialize_game()) {
				main_game_loop();
			}
		}
		numplayers = 0;
	}
#ifdef WINDOWS
	WSACleanup();
#endif
	return 0;
}
